import { useCallback } from 'react'
import { toast } from 'react-toastify'

import { bytesToHex } from '@noble/hashes/utils'
import { NDKEvent, NDKFilter, NDKKind } from '@nostr-dev-kit/ndk'
import _ from 'lodash'
import {
  Event,
  generateSecretKey,
  getPublicKey,
  kinds,
  UnsignedEvent
} from 'nostr-tools'

import { useAppDispatch, useAppSelector, useNDKContext } from '.'
import { NostrController } from '../controllers'
import {
  updateProcessedGiftWraps,
  updateUserAppData as updateUserAppDataAction
} from '../store/actions'
import { Keys } from '../store/auth/types'
import {
  isSigitNotification,
  Meta,
  SigitNotification,
  UserAppData,
  UserRelaysType
} from '../types'
import {
  countLeadingZeroes,
  createWrap,
  deleteBlossomFile,
  fetchMetaFromFileStorage,
  getDTagForUserAppData,
  getUserAppDataFromBlossom,
  hexToNpub,
  parseJson,
  unixNow,
  uploadUserAppDataToBlossom
} from '../utils'

export const useNDK = () => {
  const dispatch = useAppDispatch()
  const { ndk, fetchEvent, fetchEventsFromUserRelays, publish } =
    useNDKContext()
  const usersPubkey = useAppSelector((state) => state.auth.usersPubkey)
  const appData = useAppSelector((state) => state.userAppData)
  const processedEvents = useAppSelector(
    (state) => state.userAppData?.processedGiftWraps
  )

  /**
   * Fetches user application data based on user's public key.
   *
   * @returns The user application data or null if an error occurs or no data is found.
   */
  const getUsersAppData = useCallback(async (): Promise<UserAppData | null> => {
    if (!usersPubkey) return null

    // Generate an identifier for the user's nip78
    const dTag = await getDTagForUserAppData()
    if (!dTag) return null

    // Define a filter for fetching events
    const filter: NDKFilter = {
      kinds: [NDKKind.AppSpecificData],
      authors: [usersPubkey],
      '#d': [dTag]
    }

    const encryptedContent = await fetchEvent(filter)
      .then((event) => {
        if (event) return event.content

        // If no event is found, return an empty stringified object
        return '{}'
      })
      .catch((err) => {
        // Log error and show a toast notification if fetching event fails
        console.log(`An error occurred in finding kind 30078 event`, err)
        toast.error(
          'An error occurred in finding kind 30078 event for data storage'
        )
        return null
      })

    // Return null if encrypted content retrieval fails
    if (!encryptedContent) return null

    // Handle case where the encrypted content is an empty object
    if (encryptedContent === '{}') {
      // Generate ephemeral key pair
      const secret = generateSecretKey()
      const pubKey = getPublicKey(secret)

      return {
        sigits: {},
        processedGiftWraps: [],
        blossomUrls: [],
        keyPair: {
          private: bytesToHex(secret),
          public: pubKey
        }
      }
    }

    // Get an instance of the NostrController
    const nostrController = NostrController.getInstance()

    // Decrypt the encrypted content
    const decrypted = await nostrController
      .nip04Decrypt(usersPubkey, encryptedContent)
      .catch((err) => {
        // Log error and show a toast notification if decryption fails
        console.log('An error occurred while decrypting app data', err)
        toast.error('An error occurred while decrypting app data')
        return null
      })

    // Return null if decryption fails
    if (!decrypted) return null

    // Parse the decrypted content
    const parsedContent = await parseJson<{
      blossomUrls: string[]
      keyPair: Keys
    }>(decrypted).catch((err) => {
      // Log error and show a toast notification if parsing fails
      console.log(
        'An error occurred in parsing the content of kind 30078 event',
        err
      )
      toast.error(
        'An error occurred in parsing the content of kind 30078 event'
      )
      return null
    })

    // Return null if parsing fails
    if (!parsedContent) return null

    const { blossomUrls, keyPair } = parsedContent

    // Return null if no blossom URLs are found
    if (blossomUrls.length === 0) return null

    // Fetch additional user app data from the first blossom URL
    const dataFromBlossom = await getUserAppDataFromBlossom(
      blossomUrls[0],
      keyPair.private
    )

    // Return null if fetching data from blossom fails
    if (!dataFromBlossom) return null

    const { sigits, processedGiftWraps } = dataFromBlossom

    // Return the final user application data
    return {
      blossomUrls,
      keyPair,
      sigits,
      processedGiftWraps
    }
  }, [usersPubkey, fetchEvent])

  const updateUsersAppData = useCallback(
    async (meta: Meta) => {
      if (!appData || !appData.keyPair || !usersPubkey) return null

      const sigits = _.cloneDeep(appData.sigits)

      const createSignatureEvent = await parseJson<Event>(
        meta.createSignature
      ).catch((err) => {
        console.log('err in parsing the createSignature event:>> ', err)
        toast.error(
          err.message || 'error occurred in parsing the create signature event'
        )
        return null
      })

      if (!createSignatureEvent) return null

      const id = createSignatureEvent.id
      let isUpdated = false

      // check if sigit already exists
      if (id in sigits) {
        // update meta only if incoming meta is more recent
        // than already existing one
        const existingMeta = sigits[id]
        if (existingMeta.modifiedAt < meta.modifiedAt) {
          sigits[id] = meta
          isUpdated = true
        }
      } else {
        sigits[id] = meta
        isUpdated = true
      }

      if (!isUpdated) return null

      const blossomUrls = [...appData.blossomUrls]

      const newBlossomUrl = await uploadUserAppDataToBlossom(
        sigits,
        appData.processedGiftWraps,
        appData.keyPair.private
      ).catch((err) => {
        console.log(
          'An error occurred in uploading user app data file to blossom server',
          err
        )
        toast.error(
          'An error occurred in uploading user app data file to blossom server'
        )
        return null
      })

      if (!newBlossomUrl) return null

      // insert new blossom url at the start of the array
      blossomUrls.unshift(newBlossomUrl)

      // only keep last 10 blossom urls, delete older ones
      if (blossomUrls.length > 10) {
        const filesToDelete = blossomUrls.splice(10)
        filesToDelete.forEach((url) => {
          deleteBlossomFile(url, appData.keyPair!.private).catch((err) => {
            console.log(
              'An error occurred in removing old file of user app data from blossom server',
              err
            )
          })
        })
      }

      // encrypt content for storing in kind 30078 event
      const nostrController = NostrController.getInstance()
      const encryptedContent = await nostrController
        .nip04Encrypt(
          usersPubkey,
          JSON.stringify({
            blossomUrls,
            keyPair: appData.keyPair
          })
        )
        .catch((err) => {
          console.log(
            'An error occurred in encryption of content for app data',
            err
          )
          toast.error(
            err.message ||
              'An error occurred in encryption of content for app data'
          )
          return null
        })

      if (!encryptedContent) return null

      // generate the identifier for user's appData event
      const dTag = await getDTagForUserAppData()
      if (!dTag) return null

      const updatedEvent: UnsignedEvent = {
        kind: kinds.Application,
        pubkey: usersPubkey,
        created_at: unixNow(),
        tags: [['d', dTag]],
        content: encryptedContent
      }

      const signedEvent = await nostrController
        .signEvent(updatedEvent)
        .catch((err) => {
          console.log('An error occurred in signing event', err)
          toast.error(err.message || 'An error occurred in signing event')
          return null
        })

      if (!signedEvent) return null

      const ndkEvent = new NDKEvent(ndk, signedEvent)
      const publishResult = await publish(ndkEvent)

      if (publishResult.length === 0) {
        toast.error(
          'An unexpected error occurred in publishing updated app data '
        )
        return null
      }

      if (!publishResult) return null

      // update redux store
      dispatch(
        updateUserAppDataAction({
          sigits,
          blossomUrls,
          processedGiftWraps: [...appData.processedGiftWraps],
          keyPair: {
            ...appData.keyPair
          }
        })
      )

      return signedEvent
    },
    [appData, dispatch, ndk, publish, usersPubkey]
  )

  const processReceivedEvent = useCallback(
    async (event: NDKEvent, difficulty: number = 5) => {
      // Abort processing if userAppData is undefined
      if (!processedEvents) return

      if (processedEvents.includes(event.id)) return

      dispatch(updateProcessedGiftWraps([...processedEvents, event.id]))

      // validate PoW
      // Count the number of leading zero bits in the hash
      const leadingZeroes = countLeadingZeroes(event.id)
      if (leadingZeroes < difficulty) return

      // decrypt the content of gift wrap event
      const nostrController = NostrController.getInstance()
      const decrypted = await nostrController.nip44Decrypt(
        event.pubkey,
        event.content
      )

      const internalUnsignedEvent = await parseJson<UnsignedEvent>(
        decrypted
      ).catch((err) => {
        console.log(
          'An error occurred in parsing the internal unsigned event',
          err
        )
        return null
      })

      if (!internalUnsignedEvent || internalUnsignedEvent.kind !== 938) return

      const parsedContent = await parseJson<Meta | SigitNotification>(
        internalUnsignedEvent.content
      ).catch((err) => {
        console.log(
          'An error occurred in parsing the internal unsigned event',
          err
        )
        return null
      })

      if (!parsedContent) return

      let meta: Meta

      if (isSigitNotification(parsedContent)) {
        const notification = parsedContent
        if (!notification.keys || !usersPubkey) return

        let encryptionKey: string | undefined

        const { sender, keys } = notification.keys

        // Retrieve the user's public key from the state
        const usersNpub = hexToNpub(usersPubkey)

        // Check if the user's public key is in the keys object
        if (usersNpub in keys) {
          // Instantiate the NostrController to decrypt the encryption key
          const nostrController = NostrController.getInstance()
          const decrypted = await nostrController
            .nip04Decrypt(sender, keys[usersNpub])
            .catch((err) => {
              console.log('An error occurred in decrypting encryption key', err)
              return undefined
            })

          encryptionKey = decrypted
        }
        try {
          meta = await fetchMetaFromFileStorage(
            notification.metaUrl,
            encryptionKey
          )
        } catch (error) {
          console.error(
            `An error occured fetching meta file from storage`,
            error
          )
          return
        }
      } else {
        meta = parsedContent
      }

      await updateUsersAppData(meta)
    },
    [dispatch, processedEvents, updateUsersAppData, usersPubkey]
  )

  const subscribeForSigits = useCallback(
    async (pubkey: string) => {
      // Define the filter for the subscription
      const filter: NDKFilter = {
        kinds: [1059 as NDKKind],
        '#p': [pubkey]
      }

      // Process the received event synchronously
      const events = await fetchEventsFromUserRelays(
        filter,
        pubkey,
        UserRelaysType.Read
      )

      for (const e of events) {
        await processReceivedEvent(e)
      }
    },
    [fetchEventsFromUserRelays, processReceivedEvent]
  )

  /**
   * Function to send a notification to a specified receiver.
   * @param receiver - The recipient's public key.
   * @param notification - Url pointing to metadata associated with the notification on blossom and keys to decrypt.
   */
  const sendNotification = useCallback(
    async (receiver: string, notification: SigitNotification) => {
      if (!usersPubkey) return

      // Create an unsigned event object with the provided metadata
      const unsignedEvent: UnsignedEvent = {
        kind: 938,
        pubkey: usersPubkey,
        content: JSON.stringify(notification),
        tags: [],
        created_at: unixNow()
      }

      // Wrap the unsigned event with the receiver's information
      const wrappedEvent = createWrap(unsignedEvent, receiver)

      // Publish the notification event to the recipient's read relays
      const ndkEvent = new NDKEvent(ndk, wrappedEvent)
      await publish(ndkEvent).catch((err) => {
        // Log an error if publishing the notification event fails
        console.log(
          `An error occurred while publishing notification event for ${hexToNpub(receiver)}`,
          err
        )
        throw err
      })
    },
    [ndk, publish, usersPubkey]
  )

  return {
    getUsersAppData,
    subscribeForSigits,
    updateUsersAppData,
    sendNotification
  }
}
